/**
 * @file ex2.c
 * @author Florian Burgener
 * @brief Exercice 2
 * @version 1.0
 * @date 2022-05-03
 *
 * @copyright Copyright (c) 2021
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define KEY_MAXIMUM_LENGTH 4
#define HASH_MAP_LENGTH 10
#define VECTOR_DEFAULT_CAPACITY 100

typedef struct Entry {
    char *key;
    double value;
} Entry;

typedef struct Vector {
    int capacity;
    int length;
    Entry **items;
} Vector;

typedef struct HashMap {
    int length;
    Vector **entries;
} HashMap;

/**
 * @brief Hashes a string.
 *
 * @param key
 * @param length
 * @return int
 */
int hash(char key[4], int length);

/**
 * @brief Initializes an entry.
 *
 * @param key
 * @param value
 * @return Entry*
 */
Entry *entry_init(char *key, double value);

/**
 * @brief Frees the memory of the entry.
 *
 * @param e
 */
void entry_destroy(Entry *e);

/**
 * @brief Initializes an vector.
 *
 * @param capacity
 * @return Vector*
 */
Vector *vector_init(int capacity);

/**
 * @brief Frees the memory of the vector.
 *
 * @param e
 */
void vector_destroy(Vector *v);

/**
 * @brief Adds an element at the end of the vector.
 *
 * @param v
 * @param e
 */
void vector_push(Vector *v, Entry *e);

/**
 * @brief Displays the vector.
 *
 * @param v
 */
void vector_print(Vector *v);

/**
 * @brief Initializes an hash map.
 *
 * @param length
 * @return HashMap*
 */
HashMap *hash_map_init(int length);

/**
 * @brief Frees the memory of the hash map.
 *
 * @param dict
 */
void hash_map_destroy(HashMap *dict);

/**
 * @brief Inserts the value in the hash map.
 *
 * @param dict
 * @param key
 * @param value
 */
void hash_map_insert(HashMap *dict, char *key, double value);

/**
 * @brief Displays the hash map.
 *
 * @param dict
 */
void hash_map_print(HashMap *dict);

int hash(char key[4], int length) {
    int index = 0;
    for (int i = 0; i < 4; ++i) {
        index = (43 * index + key[i]) % length;
    }
    return index;
}

Entry *entry_init(char *key, double value) {
    Entry *e = (Entry *)malloc(sizeof(Entry));
    e->key = (char *)malloc(sizeof(char) * (KEY_MAXIMUM_LENGTH + 1));
    e->key[0] = '\0';
    strcpy(e->key, key);
    e->value = value;
    return e;
}

void entry_destroy(Entry *e) {
    free(e->key);
    free(e);
}

Vector *vector_init(int capacity) {
    Vector *v = (Vector *)malloc(sizeof(Vector));
    v->capacity = capacity;
    v->length = 0;
    v->items = (Entry **)malloc(sizeof(Entry *) * v->capacity);
    return v;
}

void vector_destroy(Vector *v) {
    for (int i = 0; i < v->length; i += 1) {
        entry_destroy(v->items[i]);
    }

    free(v->items);
    free(v);
}

void vector_push(Vector *v, Entry *e) {
    v->items[v->length] = e;
    v->length += 1;
}

void vector_print(Vector *v) {
    for (int i = 0; i < v->length; i += 1) {
        printf("(%s, %.1lf)", v->items[i]->key, v->items[i]->value);

        if (i < v->length - 1) {
            printf(", ");
        }
    }
}

HashMap *hash_map_init(int length) {
    HashMap *dict = (HashMap *)malloc(sizeof(HashMap));
    dict->length = length;
    dict->entries = (Vector **)malloc(sizeof(Vector *) * dict->length);

    for (int i = 0; i < length; i += 1) {
        dict->entries[i] = vector_init(VECTOR_DEFAULT_CAPACITY);
    }

    return dict;
}

void hash_map_destroy(HashMap *dict) {
    for (int i = 0; i < dict->length; i += 1) {
        vector_destroy(dict->entries[i]);
    }

    free(dict->entries);
    free(dict);
}

void hash_map_insert(HashMap *dict, char *key, double value) {
    int index = hash(key, dict->length);
    Vector *v = dict->entries[index];

    for (int i = 0; i < v->length; i += 1) {
        if (strcmp(v->items[i]->key, key) == 0) {
            v->items[i]->value = value;
            return;
        }
    }

    Entry *e = entry_init(key, value);
    vector_push(v, e);
}

void hash_map_print(HashMap *dict) {
    for (int i = 0; i < dict->length; i += 1) {
        printf("%d: ", i);
        vector_print(dict->entries[i]);
        printf("\n");
    }
}

int main() {
    HashMap *dict = hash_map_init(HASH_MAP_LENGTH);

    hash_map_insert(dict, "ajku", 1.1);
    hash_map_insert(dict, "ajkv", 2.2);
    hash_map_insert(dict, "ajkw", 3.1);
    hash_map_insert(dict, "ajkx", 4.2);
    hash_map_insert(dict, "ajky", 5.7);
    hash_map_insert(dict, "ajkz", 6.1);
    hash_map_insert(dict, "ajka", 1.8);
    hash_map_insert(dict, "ajkb", 2.7);
    hash_map_insert(dict, "ajkc", 3.6);
    hash_map_insert(dict, "bjkd", 1.9);
    hash_map_print(dict);

    printf("\n");

    hash_map_insert(dict, "ajkt", 0.1);
    hash_map_insert(dict, "ajku", 9.1);
    hash_map_insert(dict, "ajkx", 2.9);
    hash_map_print(dict);

    hash_map_destroy(dict);
    return EXIT_SUCCESS;
}
